import json
import logging
import textwrap
from functools import partial
from typing import Any, Awaitable, Callable, Dict, List, Optional
from unittest.mock import MagicMock

import anyio
import pytest
import yaml
from botocore.exceptions import ClientError
from moto import mock_ec2, mock_ecs, mock_logs
from moto.ec2.utils import generate_instance_identity_document
from prefect.exceptions import InfrastructureNotAvailable, InfrastructureNotFound
from prefect.logging.configuration import setup_logging
from prefect.server.schemas.core import Deployment, Flow, FlowRun
from prefect.utilities.asyncutils import run_sync_in_worker_thread
from prefect.utilities.dockerutils import get_prefect_image_name
from pydantic import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic.v1 import ValidationError
else:
    from pydantic import ValidationError

from prefect_aws.ecs import (
    ECS_DEFAULT_CPU,
    ECS_DEFAULT_MEMORY,
    ECSTask,
    get_container,
    get_prefect_container,
    parse_task_identifier,
)

setup_logging()


BASE_TASK_DEFINITION_YAML = """
containerDefinitions:
- cpu: 1024
  image: prefecthq/prefect:2.1.0-python3.8
  memory: 2048
  name: prefect
family: prefect
"""

BASE_TASK_DEFINITION = yaml.safe_load(BASE_TASK_DEFINITION_YAML)


def inject_moto_patches(moto_mock, patches: Dict[str, List[Callable]]):
    def injected_call(method, patch_list, *args, **kwargs):
        for patch in patch_list:
            result = patch(method, *args, **kwargs)
        return result

    for account in moto_mock.backends:
        for region in moto_mock.backends[account]:
            backend = moto_mock.backends[account][region]

            for attr, attr_patches in patches.items():
                original_method = getattr(backend, attr)
                setattr(
                    backend, attr, partial(injected_call, original_method, attr_patches)
                )


def patch_run_task(mock, run_task, *args, **kwargs):
    """
    Track calls to `run_task` by calling a mock as well.
    """
    mock(*args, **kwargs)
    return run_task(*args, **kwargs)


def patch_describe_tasks_add_prefect_container(describe_tasks, *args, **kwargs):
    """
    Adds the minimal prefect container to moto's task description.
    """
    result = describe_tasks(*args, **kwargs)
    for task in result:
        if not task.containers:
            task.containers = []
        if not get_prefect_container(task.containers):
            task.containers.append({"name": "prefect"})
    return result


def patch_calculate_task_resource_requirements(
    _calculate_task_resource_requirements, task_definition
):
    """
    Adds support for non-EC2 execution modes to moto's calculation of task definition.
    """
    for container_definition in task_definition.container_definitions:
        container_definition.setdefault("memory", 0)
    return _calculate_task_resource_requirements(task_definition)


def create_log_stream(session, run_task, *args, **kwargs):
    """
    When running a task, create the log group and stream if logging is configured on
    containers.

    See https://docs.aws.amazon.com/AmazonECS/latest/developerguide/using_awslogs.html
    """
    tasks = run_task(*args, **kwargs)
    if not tasks:
        return tasks
    task = tasks[0]

    ecs_client = session.client("ecs")
    logs_client = session.client("logs")

    task_definition = ecs_client.describe_task_definition(
        taskDefinition=task.task_definition_arn
    )["taskDefinition"]

    for container in task_definition.get("containerDefinitions", []):
        log_config = container.get("logConfiguration", {})
        if log_config:
            if log_config.get("logDriver") != "awslogs":
                continue

            options = log_config.get("options", {})
            if not options:
                raise ValueError("logConfiguration does not include options.")

            group_name = options.get("awslogs-group")
            if not group_name:
                raise ValueError(
                    "logConfiguration.options does not include awslogs-group"
                )

            if options.get("awslogs-create-group") == "true":
                logs_client.create_log_group(logGroupName=group_name)

            stream_prefix = options.get("awslogs-stream-prefix")
            if not stream_prefix:
                raise ValueError(
                    "logConfiguration.options does not include awslogs-stream-prefix"
                )

            logs_client.create_log_stream(
                logGroupName=group_name,
                logStreamName=f"{stream_prefix}/{container['name']}/{task.id}",
            )

    return tasks


def add_ec2_instance_to_ecs_cluster(session, cluster_name):
    ecs_client = session.client("ecs")
    ec2_client = session.client("ec2")
    ec2_resource = session.resource("ec2")

    ecs_client.create_cluster(clusterName=cluster_name)

    images = ec2_client.describe_images()
    image_id = images["Images"][0]["ImageId"]

    test_instance = ec2_resource.create_instances(
        ImageId=image_id, MinCount=1, MaxCount=1
    )[0]

    ecs_client.register_container_instance(
        cluster=cluster_name,
        instanceIdentityDocument=json.dumps(
            generate_instance_identity_document(test_instance)
        ),
    )


def create_test_ecs_cluster(ecs_client, cluster_name) -> str:
    """
    Create an ECS cluster and return its ARN
    """
    return ecs_client.create_cluster(clusterName=cluster_name)["cluster"]["clusterArn"]


def describe_task(ecs_client, task_arn, **kwargs) -> dict:
    """
    Describe a single ECS task
    """
    return ecs_client.describe_tasks(tasks=[task_arn], include=["TAGS"], **kwargs)[
        "tasks"
    ][0]


async def stop_task(ecs_client, task_arn, **kwargs):
    """
    Stop an ECS task.

    Additional keyword arguments are passed to `ECSClient.stop_task`.
    """
    task = await run_sync_in_worker_thread(describe_task, ecs_client, task_arn)
    # Check that the task started successfully
    assert task["lastStatus"] == "RUNNING", "Task should be RUNNING before stopping"
    print("Stopping task...")
    await run_sync_in_worker_thread(ecs_client.stop_task, task=task_arn, **kwargs)


def describe_task_definition(ecs_client, task):
    return ecs_client.describe_task_definition(
        taskDefinition=task["taskDefinitionArn"]
    )["taskDefinition"]


async def run_then_stop_task(
    task: ECSTask, after_start: Optional[Callable[[str], Awaitable[Any]]] = None
) -> str:
    """
    Run an ECS Task then stop it.

    Moto will not advance the state of tasks, so `ECSTask.run` would hang forever if
    the run is created successfully and not stopped.

    `after_start` can be used to run something after the task starts but before it is
    stopped. It will be passed the task arn.
    """
    session = task.aws_credentials.get_boto3_session()

    with anyio.fail_after(20):
        async with anyio.create_task_group() as tg:
            identifier = await tg.start(task.run)
            cluster, task_arn = parse_task_identifier(identifier)

            if after_start:
                await after_start(task_arn)

            # Stop the task after it starts to prevent the test from running forever
            tg.start_soon(
                partial(stop_task, session.client("ecs"), task_arn, cluster=cluster)
            )

    return task_arn


@pytest.fixture(autouse=True)
def patch_task_watch_poll_interval(monkeypatch):
    # Patch the poll interval to be way shorter for speed during testing!
    monkeypatch.setattr(ECSTask.__fields__["task_watch_poll_interval"], "default", 0.05)


@pytest.fixture
def ecs_mocks(aws_credentials):
    with mock_ecs() as ecs:
        with mock_ec2():
            with mock_logs():
                session = aws_credentials.get_boto3_session()

                inject_moto_patches(
                    ecs,
                    {
                        # Ensure container is created in described tasks
                        "describe_tasks": [patch_describe_tasks_add_prefect_container],
                        # Fix moto internal resource requirement calculations
                        "_calculate_task_resource_requirements": [
                            patch_calculate_task_resource_requirements
                        ],
                        # Add log group creation
                        "run_task": [partial(create_log_stream, session)],
                    },
                )

                create_test_ecs_cluster(session.client("ecs"), "default")

                # NOTE: Even when using FARGATE, moto requires container instances to be
                #       registered. This differs from AWS behavior.
                add_ec2_instance_to_ecs_cluster(session, "default")

                yield ecs


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("launch_type", ["EC2", "FARGATE", "FARGATE_SPOT"])
async def test_launch_types(aws_credentials, launch_type: str):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["prefect", "version"],
        launch_type=launch_type,
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    if launch_type != "FARGATE_SPOT":
        assert launch_type in task_definition["compatibilities"]
        assert task["launchType"] == launch_type
    else:
        assert "FARGATE" in task_definition["compatibilities"]
        # FARGATE SPOT requires a null launch type
        assert not task.get("launchType")
        # Instead, it requires a capacity provider strategy but this is not supported
        # by moto and is not present on the task even when provided
        # assert task["capacityProviderStrategy"] == [
        #     {"capacityProvider": "FARGATE_SPOT", "weight": 1}
        # ]

    requires_capabilities = task_definition.get("requiresCompatibilities", [])
    if launch_type != "EC2":
        assert "FARGATE" in requires_capabilities
    else:
        assert not requires_capabilities


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("launch_type", ["EC2", "FARGATE", "FARGATE_SPOT"])
@pytest.mark.parametrize(
    "cpu,memory", [(None, None), (1024, None), (None, 2048), (2048, 4096)]
)
async def test_cpu_and_memory(aws_credentials, launch_type: str, cpu: int, memory: int):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["prefect", "version"],
        launch_type=launch_type,
        cpu=cpu,
        memory=memory,
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    container_definition = get_prefect_container(
        task_definition["containerDefinitions"]
    )
    overrides = task["overrides"]
    container_overrides = get_prefect_container(overrides["containerOverrides"])

    if launch_type == "EC2":
        # EC2 requires CPU and memory to be defined at the container level
        assert container_definition["cpu"] == cpu or ECS_DEFAULT_CPU
        assert container_definition["memory"] == memory or ECS_DEFAULT_MEMORY
    else:
        # Fargate requires CPU and memory to be defined at the task definition level
        assert task_definition["cpu"] == str(cpu or ECS_DEFAULT_CPU)
        assert task_definition["memory"] == str(memory or ECS_DEFAULT_MEMORY)

    # We always provide non-null values as overrides on the task run
    assert overrides.get("cpu") == (str(cpu) if cpu else None)
    assert overrides.get("memory") == (str(memory) if memory else None)
    # And as overrides for the Prefect container
    assert container_overrides.get("cpu") == cpu
    assert container_overrides.get("memory") == memory


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("launch_type", ["EC2", "FARGATE", "FARGATE_SPOT"])
async def test_network_mode_default(aws_credentials, launch_type: str):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["prefect", "version"],
        launch_type=launch_type,
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    if launch_type == "EC2":
        assert task_definition["networkMode"] == "bridge"
    else:
        assert task_definition["networkMode"] == "awsvpc"


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("launch_type", ["EC2", "FARGATE", "FARGATE_SPOT"])
async def test_container_command(aws_credentials, launch_type: str):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["prefect", "version"],
        launch_type=launch_type,
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)

    container_overrides = get_prefect_container(task["overrides"]["containerOverrides"])
    assert container_overrides["command"] == ["prefect", "version"]


@pytest.mark.usefixtures("ecs_mocks")
async def test_environment_variables(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        env={"FOO": "BAR"},
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    prefect_container_definition = get_prefect_container(
        task_definition["containerDefinitions"]
    )
    assert not prefect_container_definition[
        "environment"
    ], "Variables should not be passed until runtime"

    prefect_container_overrides = get_prefect_container(
        task["overrides"]["containerOverrides"]
    )
    expected = [
        {"name": key, "value": value}
        for key, value in ECSTask._base_environment().items()
    ]
    expected.append({"name": "FOO", "value": "BAR"})
    assert prefect_container_overrides.get("environment") == expected


@pytest.mark.usefixtures("ecs_mocks")
async def test_labels(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        labels={"foo": "bar"},
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    assert not task_definition.get("tags"), "Labels should not be passed until runtime"

    assert task.get("tags") == [{"key": "foo", "value": "bar"}]


@pytest.mark.usefixtures("ecs_mocks")
async def test_container_command_from_task_definition(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition={
            "containerDefinitions": [{"name": "prefect", "command": ["echo", "hello"]}]
        },
        command=[],
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)

    container_overrides = get_prefect_container(task["overrides"]["containerOverrides"])
    assert "command" not in container_overrides


@pytest.mark.usefixtures("ecs_mocks")
async def test_extra_containers_in_task_definition(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition={
            "containerDefinitions": [
                {"name": "secondary", "command": ["echo", "hello"], "image": "alpine"}
            ]
        },
        command=["prefect", "version"],
        image="test",
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    user_container = get_container(task_definition["containerDefinitions"], "secondary")
    assert (
        user_container is not None
    ), "The user-specified container should be present still"
    assert user_container["command"] == ["echo", "hello"]
    assert user_container["image"] == "alpine", "The image should be left unchanged"

    prefect_container = get_prefect_container(task_definition["containerDefinitions"])
    assert prefect_container is not None, "The prefect container should be added"
    assert (
        prefect_container["image"] == "test"
    ), "The prefect container should use the image field"

    container_overrides = task["overrides"]["containerOverrides"]
    user_container_overrides = get_container(container_overrides, "secondary")
    prefect_container_overrides = get_prefect_container(container_overrides)
    assert (
        user_container_overrides is None
    ), "The user container should not be included in overrides"
    assert (
        prefect_container_overrides
    ), "The prefect container should have overrides still"


@pytest.mark.usefixtures("ecs_mocks")
async def test_prefect_container_in_task_definition(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition={
            "containerDefinitions": [
                {
                    "name": "prefect",
                    "command": ["should", "be", "gone"],
                    "image": "should-be-gone",
                    "privileged": True,
                }
            ]
        },
        command=["prefect", "version"],
        image="test",
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    prefect_container = get_prefect_container(task_definition["containerDefinitions"])

    assert (
        prefect_container["image"] == "test"
    ), "The prefect container should use the image field"

    assert prefect_container["command"] == [
        "should",
        "be",
        "gone",
    ], "The command should be left unchanged on the task definition"

    assert (
        prefect_container["privileged"] is True
    ), "Extra attributes should be retained"

    container_overrides = get_prefect_container(task["overrides"]["containerOverrides"])
    assert container_overrides["command"] == [
        "prefect",
        "version",
    ], "The command should be passed as an override"


@pytest.mark.usefixtures("ecs_mocks")
async def test_default_image_in_task_definition(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition={
            "containerDefinitions": [
                {
                    "name": "prefect",
                    "image": "use-this-image",
                }
            ]
        },
        command=["prefect", "version"],
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    # The image on the block is inferred from the task defintinion
    assert task.image == "use-this-image"

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    prefect_container = get_prefect_container(task_definition["containerDefinitions"])
    assert (
        prefect_container["image"] == "use-this-image"
    ), "The image from the task definition should be used"


@pytest.mark.usefixtures("ecs_mocks")
async def test_image_overrides_task_definition(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition={
            "containerDefinitions": [
                {
                    "name": "prefect",
                    "image": "use-this-image",
                }
            ]
        },
        command=["prefect", "version"],
        image="override-image",
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    prefect_container = get_prefect_container(task_definition["containerDefinitions"])
    assert (
        prefect_container["image"] == "override-image"
    ), "The provided image should override task definition"


@pytest.mark.parametrize(
    "task_definition",
    [
        # Empty task definition
        {},
        # Task definition with prefect container but no image
        {
            "containerDefinitions": [
                {
                    "name": "prefect",
                }
            ]
        },
        # Task definition with other container with image
        {"containerDefinitions": [{"name": "foo", "image": "not-me-image"}]},
    ],
)
@pytest.mark.usefixtures("ecs_mocks")
async def test_default_image(aws_credentials, task_definition):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition=task_definition,
        command=["prefect", "version"],
    )
    print(task.preview())
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    # The image on the block is inferred from Prefect/Python versions
    assert task.image == get_prefect_image_name()

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    prefect_container = get_prefect_container(task_definition["containerDefinitions"])
    assert (
        prefect_container["image"] == get_prefect_image_name()
    ), "The image should be the default Prefect tag"


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("launch_type", ["EC2", "FARGATE", "FARGATE_SPOT"])
async def test_default_cpu_and_memory_in_task_definition(
    aws_credentials, launch_type: str
):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition={
            "containerDefinitions": [
                {
                    "name": "prefect",
                    "command": ["should", "be", "gone"],
                    "image": "should-be-gone",
                    "cpu": 2048,
                    "memory": 4096,
                }
            ],
            "cpu": "4096",
            "memory": "8192",
        },
        command=["prefect", "version"],
        image="test",
        launch_type=launch_type,
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    container_definition = get_prefect_container(
        task_definition["containerDefinitions"]
    )
    overrides = task["overrides"]
    container_overrides = get_prefect_container(overrides["containerOverrides"])

    # All of these values should be retained
    assert container_definition["cpu"] == 2048
    assert container_definition["memory"] == 4096
    assert task_definition["cpu"] == str(4096)
    assert task_definition["memory"] == str(8192)

    # No values should be overriden at runtime
    assert overrides.get("cpu") is None
    assert overrides.get("memory") is None
    assert container_overrides.get("cpu") is None
    assert container_overrides.get("memory") is None


@pytest.mark.usefixtures("ecs_mocks")
async def test_environment_variables_in_task_definition(aws_credentials):
    # See also, `test_unset_environment_variables_in_task_definition`
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition={
            "containerDefinitions": [
                {
                    "name": "prefect",
                    "environment": [
                        {"name": "BAR", "value": "FOO"},
                        {"name": "OVERRIDE", "value": "OLD"},
                    ],
                }
            ],
        },
        env={"FOO": "BAR", "OVERRIDE": "NEW"},
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    prefect_container_definition = get_prefect_container(
        task_definition["containerDefinitions"]
    )

    assert prefect_container_definition["environment"] == [
        {"name": "BAR", "value": "FOO"},
        {"name": "OVERRIDE", "value": "OLD"},
    ]

    prefect_container_overrides = get_prefect_container(
        task["overrides"]["containerOverrides"]
    )
    expected_base = [
        {"name": key, "value": value}
        for key, value in ECSTask._base_environment().items()
    ]
    assert prefect_container_overrides.get("environment") == expected_base + [
        {"name": "FOO", "value": "BAR"},
        {"name": "OVERRIDE", "value": "NEW"},
    ]


@pytest.mark.usefixtures("ecs_mocks")
async def test_unset_environment_variables_in_task_definition(aws_credentials):
    # In contrast to `test_environment_variables_in_task_definition`, this tests the
    # use of `None` in `ECSTask.env` values to signal _removal_ of an environment
    # variable instead of overriding a value.
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition={
            "containerDefinitions": [
                {
                    "name": "prefect",
                    "environment": [
                        {"name": "FOO", "value": "FOO"},
                        {"name": "BAR", "value": "BAR"},
                    ],
                }
            ]
        },
        env={"FOO": None},
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    prefect_container_definition = get_prefect_container(
        task_definition["containerDefinitions"]
    )
    assert prefect_container_definition["environment"] == [
        {"name": "BAR", "value": "BAR"}
    ], "FOO should be removed from the task definition"

    expected_base = [
        {"name": key, "value": value}
        for key, value in ECSTask._base_environment().items()
    ]
    prefect_container_overrides = get_prefect_container(
        task["overrides"]["containerOverrides"]
    )
    assert (
        prefect_container_overrides.get("environment") == expected_base
    ), "FOO should not be passed at runtime"


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("provided_as_field", [True, False])
async def test_execution_role_arn_in_task_definition(
    aws_credentials, provided_as_field: bool
):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition={"executionRoleArn": "test"},
        execution_role_arn="override" if provided_as_field else None,
    )
    print(task.preview())

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    # Check if it is overidden if provided as a field
    assert (
        task_definition["executionRoleArn"] == "test"
        if not provided_as_field
        else "override"
    )


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("default_cluster", [True, False])
async def test_cluster(aws_credentials, default_cluster: bool):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    # Construct a non-default cluster. We build this in either case since otherwise
    # there is only one cluster and there's no choice but to use the default.
    second_cluster_arn = create_test_ecs_cluster(ecs_client, "second-cluster")
    add_ec2_instance_to_ecs_cluster(session, "second-cluster")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        cluster=None if default_cluster else "second-cluster",
    )
    print(task.preview())

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)

    if default_cluster:
        assert task["clusterArn"].endswith("default")
    else:
        assert task["clusterArn"] == second_cluster_arn


@pytest.mark.usefixtures("ecs_mocks")
async def test_execution_role_arn(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        execution_role_arn="test",
    )
    print(task.preview())

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    assert task_definition["executionRoleArn"] == "test"


@pytest.mark.usefixtures("ecs_mocks")
async def test_task_role_arn(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_role_arn="test",
    )
    print(task.preview())

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)

    assert task["overrides"]["taskRoleArn"] == "test"


@pytest.mark.usefixtures("ecs_mocks")
async def test_network_config_from_vpc_id(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ec2_resource = session.resource("ec2")
    vpc = ec2_resource.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2_resource.create_subnet(CidrBlock="10.0.2.0/24", VpcId=vpc.id)

    task = ECSTask(aws_credentials=aws_credentials, vpc_id=vpc.id)

    # Capture the task run call because moto does not track 'networkConfiguration'
    original_run_task = task._run_task
    mock_run_task = MagicMock(side_effect=original_run_task)
    task._run_task = mock_run_task

    print(task.preview())

    await run_then_stop_task(task)

    network_configuration = mock_run_task.call_args[0][1].get("networkConfiguration")

    # Subnet ids are copied from the vpc
    assert network_configuration == {
        "awsvpcConfiguration": {
            "subnets": [subnet.id],
            "assignPublicIp": "ENABLED",
            "securityGroups": [],
        }
    }


@pytest.mark.usefixtures("ecs_mocks")
async def test_network_config_from_default_vpc(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ec2_client = session.client("ec2")

    default_vpc_id = ec2_client.describe_vpcs(
        Filters=[{"Name": "isDefault", "Values": ["true"]}]
    )["Vpcs"][0]["VpcId"]
    default_subnets = ec2_client.describe_subnets(
        Filters=[{"Name": "vpc-id", "Values": [default_vpc_id]}]
    )["Subnets"]

    task = ECSTask(aws_credentials=aws_credentials)

    # Capture the task run call because moto does not track 'networkConfiguration'
    original_run_task = task._run_task
    mock_run_task = MagicMock(side_effect=original_run_task)
    task._run_task = mock_run_task

    print(task.preview())

    await run_then_stop_task(task)

    network_configuration = mock_run_task.call_args[0][1].get("networkConfiguration")

    # Subnet ids are copied from the vpc
    assert network_configuration == {
        "awsvpcConfiguration": {
            "subnets": [subnet["SubnetId"] for subnet in default_subnets],
            "assignPublicIp": "ENABLED",
            "securityGroups": [],
        }
    }


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("explicit_network_mode", [True, False])
async def test_network_config_is_empty_without_awsvpc_network_mode(
    aws_credentials, explicit_network_mode
):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        # EC2 uses the 'bridge' network mode by default but we want to have test
        # coverage for when it is set on the task definition
        task_definition={"networkMode": "bridge"} if explicit_network_mode else None,
        # FARGATE requires the 'awsvpc' network mode
        launch_type="EC2",
    )

    # Capture the task run call because moto does not track 'networkConfiguration'
    original_run_task = task._run_task
    mock_run_task = MagicMock(side_effect=original_run_task)
    task._run_task = mock_run_task

    print(task.preview())

    await run_then_stop_task(task)

    network_configuration = mock_run_task.call_args[0][1].get("networkConfiguration")
    assert network_configuration is None


@pytest.mark.usefixtures("ecs_mocks")
async def test_network_config_missing_default_vpc(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ec2_client = session.client("ec2")

    default_vpc_id = ec2_client.describe_vpcs(
        Filters=[{"Name": "isDefault", "Values": ["true"]}]
    )["Vpcs"][0]["VpcId"]
    ec2_client.delete_vpc(VpcId=default_vpc_id)

    task = ECSTask(aws_credentials=aws_credentials)

    with pytest.raises(ValueError, match="Failed to find the default VPC"):
        await run_then_stop_task(task)


@pytest.mark.usefixtures("ecs_mocks")
async def test_network_config_from_vpc_with_no_subnets(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ec2_resource = session.resource("ec2")
    vpc = ec2_resource.create_vpc(CidrBlock="172.16.0.0/16")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        vpc_id=vpc.id,
    )
    print(task.preview())

    with pytest.raises(
        ValueError, match=f"Failed to find subnets for VPC with ID {vpc.id}"
    ):
        await run_then_stop_task(task)


@pytest.mark.usefixtures("ecs_mocks")
async def test_logging_requires_execution_role_arn(aws_credentials):
    with pytest.raises(
        ValidationError,
        match="`execution_role_arn` must be provided",
    ):
        ECSTask(
            aws_credentials=aws_credentials,
            command=["prefect", "version"],
            configure_cloudwatch_logs=True,
        )


@pytest.mark.usefixtures("ecs_mocks")
async def test_log_options_requires_logging(aws_credentials):
    with pytest.raises(
        ValidationError,
        match=(  # noqa
            "`configure_cloudwatch_log` must be enabled to use"
            " `cloudwatch_logs_options`"
        ),
    ):
        ECSTask(
            aws_credentials=aws_credentials,
            command=["prefect", "version"],
            configure_cloudwatch_logs=False,
            cloudwatch_logs_options={"foo": " bar"},
        )


@pytest.mark.usefixtures("ecs_mocks")
async def test_logging_requires_execution_role_arn_at_runtime(aws_credentials):
    # In constrast to `test_logging_requires_execution_role_arn`, a task definition
    # has been provided by ARN reference and we do not know if the execution role is
    # missing until runtime.

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")
    task_definition_arn = ecs_client.register_task_definition(**BASE_TASK_DEFINITION)[
        "taskDefinition"
    ]["taskDefinitionArn"]

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["prefect", "version"],
        configure_cloudwatch_logs=True,
        task_definition_arn=task_definition_arn,
        # This test is launch type agnostic but the task definition we register receives
        # the default network mode type of 'bridge' which is not compatible with FARGATE
        launch_type="EC2",
    )
    with pytest.raises(ValueError, match="An execution role arn must be set"):
        await task.run()


@pytest.mark.usefixtures("ecs_mocks")
async def test_configure_cloudwatch_logging(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["prefect", "version"],
        configure_cloudwatch_logs=True,
        execution_role_arn="test",
    )

    task_arn = await run_then_stop_task(task)
    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    for container in task_definition["containerDefinitions"]:
        if container["name"] == "prefect":
            # Assert that the 'prefect' container has logging configured
            assert container["logConfiguration"] == {
                "logDriver": "awslogs",
                "options": {
                    "awslogs-create-group": "true",
                    "awslogs-group": "prefect",
                    "awslogs-region": "us-east-1",
                    "awslogs-stream-prefix": "prefect",
                },
            }
        else:
            # Other containers should not be modifed
            assert "logConfiguration" not in container


@pytest.mark.usefixtures("ecs_mocks")
async def test_stream_output(aws_credentials, caplog):
    session = aws_credentials.get_boto3_session()
    logs_client = session.client("logs")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["prefect", "version"],
        configure_cloudwatch_logs=True,
        stream_output=True,
        execution_role_arn="test",
        # Override the family so it does not match the container name
        family="test-family",
        # Override the prefix so it does not match the container name
        cloudwatch_logs_options={"awslogs-stream-prefix": "test-prefix"},
        # Avoid slow polling during the test
        task_watch_poll_interval=0.1,
    )

    async def write_fake_log(task_arn):
        # TODO: moto does not appear to support actually reading these logs
        #       as they do not appear during `get_log_event` calls
        # prefix/container-name/task-id
        stream_name = f"test-prefix/prefect/{task_arn.rsplit('/')[-1]}"
        logs_client.put_log_events(
            logGroupName="prefect",
            logStreamName=stream_name,
            logEvents=[
                {"timestamp": i, "message": f"test-message-{i}"} for i in range(100)
            ],
        )

    await run_then_stop_task(task, after_start=write_fake_log)

    logs_client = session.client("logs")
    streams = logs_client.describe_log_streams(logGroupName="prefect")["logStreams"]

    assert len(streams) == 1

    # Ensure we did not encounter any logging errors
    assert "Failed to read log events" not in caplog.text

    # TODO: When moto supports reading logs, fix this
    # out, err = capsys.readouterr()
    # assert "test-message-{i}" in err


@pytest.mark.usefixtures("ecs_mocks")
async def test_cloudwatch_log_options(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["prefect", "version"],
        configure_cloudwatch_logs=True,
        execution_role_arn="test",
        cloudwatch_logs_options={
            "awslogs-stream-prefix": "override-prefix",
            "max-buffer-size": "2m",
        },
    )

    task_arn = await run_then_stop_task(task)
    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)

    for container in task_definition["containerDefinitions"]:
        if container["name"] == "prefect":
            # Assert that the 'prefect' container has logging configured with user
            # provided options
            assert container["logConfiguration"] == {
                "logDriver": "awslogs",
                "options": {
                    "awslogs-create-group": "true",
                    "awslogs-group": "prefect",
                    "awslogs-region": "us-east-1",
                    "awslogs-stream-prefix": "override-prefix",
                    "max-buffer-size": "2m",
                },
            }
        else:
            # Other containers should not be modifed
            assert "logConfiguration" not in container


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("launch_type", ["FARGATE", "FARGATE_SPOT"])
async def test_bridge_network_mode_warns_on_fargate(aws_credentials, launch_type: str):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["prefect", "version"],
        task_definition={"networkMode": "bridge"},
        launch_type=launch_type,
    )
    with pytest.warns(
        UserWarning,
        match=(
            "Found network mode 'bridge' which is not compatible with launch type "
            f"{launch_type!r}"
        ),
    ):
        with pytest.raises(ClientError):
            await run_then_stop_task(task)


@pytest.mark.usefixtures("ecs_mocks")
async def test_deregister_task_definition(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=True,
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    assert task_definition["status"] == "INACTIVE"


@pytest.mark.usefixtures("ecs_mocks")
async def test_latest_task_definition_used_if_equal(aws_credentials):
    task = ECSTask(aws_credentials=aws_credentials)
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn_1 = await run_then_stop_task(task)
    task_arn_2 = await run_then_stop_task(task)

    task_1 = describe_task(ecs_client, task_arn_1)
    task_2 = describe_task(ecs_client, task_arn_2)

    assert task_1["taskDefinitionArn"] == task_2["taskDefinitionArn"]


@pytest.mark.usefixtures("ecs_mocks")
async def test_latest_task_definition_not_used_if_in_another_family(
    aws_credentials,
):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_1 = ECSTask(aws_credentials=aws_credentials, family="test1")
    task_2 = ECSTask(aws_credentials=aws_credentials, family="test2")

    task_arn_1 = await run_then_stop_task(task_1)
    task_arn_2 = await run_then_stop_task(task_2)

    task_1 = describe_task(ecs_client, task_arn_1)
    task_2 = describe_task(ecs_client, task_arn_2)

    assert task_1["taskDefinitionArn"] != task_2["taskDefinitionArn"]


@pytest.mark.usefixtures("ecs_mocks")
async def test_latest_task_definition_not_used_if_inequal(
    aws_credentials,
):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    # Place it in the same family
    task_1 = ECSTask(
        aws_credentials=aws_credentials,
        family="test",
        image="image1",
        auto_deregister_task_definition=False,
    )
    task_2 = ECSTask(
        aws_credentials=aws_credentials,
        family="test",
        image="image2",
        auto_deregister_task_definition=False,
    )

    task_arn_1 = await run_then_stop_task(task_1)
    task_arn_2 = await run_then_stop_task(task_2)

    task_1 = describe_task(ecs_client, task_arn_1)
    task_2 = describe_task(ecs_client, task_arn_2)

    assert task_1["taskDefinitionArn"] != task_2["taskDefinitionArn"]


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("launch_type", ["EC2", "FARGATE"])
@pytest.mark.parametrize(
    "overrides",
    [
        {"env": {"FOO": "BAR"}},
        {"command": ["test"]},
        {"labels": {"FOO": "BAR"}},
        {"stream_output": True, "configure_cloudwatch_logs": False},
        {"cluster": "test"},
        {"task_role_arn": "test"},
        # Note: null environment variables can cause override, but not when missing
        # from the base task definition
        {"env": {"FOO": None}},
        # The following would not result in a copy when using a task_definition_arn
        # but will be eagerly set on the new task definition and result in a cache miss
        # {"cpu": 2048},
        # {"memory": 4096},
        # {"execution_role_arn": "test"},
        # {"launch_type": "EXTERNAL"},
    ],
    ids=lambda item: str(sorted(list(set(item.keys())))),
)
async def test_latest_task_definition_with_overrides_that_do_not_require_copy(
    aws_credentials, overrides, launch_type
):
    """
    Any of these overrides should be configured at runtime and not require a new
    task definition to be registered
    """
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    if "cluster" in overrides:
        create_test_ecs_cluster(ecs_client, overrides["cluster"])
        add_ec2_instance_to_ecs_cluster(session, overrides["cluster"])

    task_1 = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        family="test",
        launch_type=launch_type,
    )
    task_2 = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        family="test",
        launch_type=launch_type,
        **overrides,
    )
    task_arn_1 = await run_then_stop_task(task_1)
    task_arn_2 = await run_then_stop_task(task_2)

    task_1 = describe_task(ecs_client, task_arn_1)
    task_2 = describe_task(ecs_client, task_arn_2)
    assert (
        task_1["taskDefinitionArn"] == task_2["taskDefinitionArn"]
    ), "The existing task definition should be used"


@pytest.mark.usefixtures("ecs_mocks")
async def test_task_definition_arn(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_definition_arn = ecs_client.register_task_definition(**BASE_TASK_DEFINITION)[
        "taskDefinition"
    ]["taskDefinitionArn"]

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition_arn=task_definition_arn,
        launch_type="EC2",
        image=None,
    )
    print(task.preview())
    task_arn = await run_then_stop_task(task)

    assert task.image is None, "Image option can be null when using task definition arn"

    task = describe_task(ecs_client, task_arn)
    assert (
        task["taskDefinitionArn"] == task_definition_arn
    ), "The task definition should be used without registering a new one"


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize(
    "overrides",
    [{"image": "new-image"}, {"configure_cloudwatch_logs": True}, {"family": "foobar"}],
)
async def test_task_definition_arn_with_overrides_that_require_copy(
    aws_credentials, overrides, caplog
):
    """
    Any of these overrides should cause the task definition to be copied and
    registered as a new version
    """
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_definition_arn = ecs_client.register_task_definition(
        **BASE_TASK_DEFINITION, executionRoleArn="base"
    )["taskDefinition"]["taskDefinitionArn"]

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition_arn=task_definition_arn,
        launch_type="EC2",
        **overrides,
    )
    print(task.preview())
    with caplog.at_level(logging.INFO, logger=task.logger.name):
        task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    assert (
        task["taskDefinitionArn"] != task_definition_arn
    ), "A new task definition should be registered"

    assert (
        "Settings require changes to the linked task definition. "
        "A new task definition will be registered. "
        "Enable DEBUG level logs to see the difference."
        in caplog.text
    )


@pytest.mark.usefixtures("ecs_mocks")
async def test_task_definition_arn_with_overrides_requiring_copy_shows_diff(
    aws_credentials, caplog
):
    """
    Any of these overrides should cause the task definition to be copied and
    registered as a new version
    """
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_definition_arn = ecs_client.register_task_definition(
        **BASE_TASK_DEFINITION, executionRoleArn="base"
    )["taskDefinition"]["taskDefinitionArn"]

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition_arn=task_definition_arn,
        launch_type="EC2",
        image="foobar",
    )
    print(task.preview())
    with caplog.at_level(logging.DEBUG, logger=task.logger.name):
        await run_then_stop_task(task)

    assert (
        "Settings require changes to the linked task definition. "
        "A new task definition will be registered. "
        in caplog.text
    )

    assert "Enable DEBUG level logs to see the difference." not in caplog.text

    expected_diff = textwrap.dedent("""
        -                            'image': 'prefecthq/prefect:2.1.0-python3.8',
        +                            'image': 'foobar',
        """)
    assert expected_diff in caplog.text


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize(
    "overrides",
    [
        {"env": {"FOO": "BAR"}},
        {"command": ["test"]},
        {"labels": {"FOO": "BAR"}},
        {"cpu": 2048},
        {"memory": 4096},
        {"execution_role_arn": "test"},
        {"stream_output": True, "configure_cloudwatch_logs": False},
        {"launch_type": "EXTERNAL"},
        {"cluster": "test"},
        {"task_role_arn": "test"},
        # Note: null environment variables can cause override, but not when missing
        # from the base task definition
        {"env": {"FOO": None}},
    ],
    ids=lambda item: str(sorted(list(set(item.keys())))),
)
async def test_task_definition_arn_with_overrides_that_do_not_require_copy(
    aws_credentials, overrides
):
    """
    Any of these overrides should be configured at runtime and not require a new
    task definition to be registered
    """
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    if "cluster" in overrides:
        create_test_ecs_cluster(ecs_client, overrides["cluster"])
        add_ec2_instance_to_ecs_cluster(session, overrides["cluster"])

    task_definition_arn = ecs_client.register_task_definition(
        **BASE_TASK_DEFINITION,
    )[
        "taskDefinition"
    ]["taskDefinitionArn"]

    # Set the default launch type for compatibility with the base task definition
    overrides.setdefault("launch_type", "EC2")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition_arn=task_definition_arn,
        image=None,
        **overrides,
    )
    print(task.preview())
    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    assert (
        task["taskDefinitionArn"] == task_definition_arn
    ), "The existing task definition should be used"


@pytest.mark.usefixtures("ecs_mocks")
async def test_deregister_task_definition_does_not_apply_to_linked_arn(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_definition_arn = ecs_client.register_task_definition(**BASE_TASK_DEFINITION)[
        "taskDefinition"
    ]["taskDefinitionArn"]

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=True,
        task_definition_arn=task_definition_arn,
        launch_type="EC2",
        image=None,
    )
    print(task.preview())
    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    describe_task_definition(ecs_client, task)["status"] == "ACTIVE"


@pytest.mark.usefixtures("ecs_mocks")
async def test_adding_security_groups_to_network_config(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ec2_resource = session.resource("ec2")
    vpc = ec2_resource.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2_resource.create_subnet(CidrBlock="10.0.2.0/24", VpcId=vpc.id)
    ec2_client = session.client("ec2")
    security_group_id = ec2_client.create_security_group(
        GroupName="test", Description="testing"
    )["GroupId"]

    task = ECSTask(
        aws_credentials=aws_credentials,
        vpc_id=vpc.id,
        task_customizations=[
            {
                "op": "add",
                "path": "/networkConfiguration/awsvpcConfiguration/securityGroups",
                "value": [security_group_id],
            },
        ],
    )

    # Capture the task run call because moto does not track 'networkConfiguration'
    original_run_task = task._run_task
    mock_run_task = MagicMock(side_effect=original_run_task)
    task._run_task = mock_run_task

    print(task.preview())

    await run_then_stop_task(task)

    network_configuration = mock_run_task.call_args[0][1].get("networkConfiguration")

    # Subnet ids are copied from the vpc
    assert network_configuration == {
        "awsvpcConfiguration": {
            "subnets": [subnet.id],
            "assignPublicIp": "ENABLED",
            "securityGroups": [security_group_id],
        }
    }


@pytest.mark.usefixtures("ecs_mocks")
async def test_disable_public_ip_in_network_config(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ec2_resource = session.resource("ec2")
    vpc = ec2_resource.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2_resource.create_subnet(CidrBlock="10.0.2.0/24", VpcId=vpc.id)

    task = ECSTask(
        aws_credentials=aws_credentials,
        vpc_id=vpc.id,
        task_customizations=[
            {
                "op": "replace",
                "path": "/networkConfiguration/awsvpcConfiguration/assignPublicIp",
                "value": "DISABLED",
            },
        ],
    )

    # Capture the task run call because moto does not track 'networkConfiguration'
    original_run_task = task._run_task
    mock_run_task = MagicMock(side_effect=original_run_task)
    task._run_task = mock_run_task

    print(task.preview())

    await run_then_stop_task(task)

    network_configuration = mock_run_task.call_args[0][1].get("networkConfiguration")

    # Subnet ids are copied from the vpc
    assert network_configuration == {
        "awsvpcConfiguration": {
            "subnets": [subnet.id],
            "assignPublicIp": "DISABLED",
            "securityGroups": [],
        }
    }


@pytest.mark.usefixtures("ecs_mocks")
async def test_custom_subnets_in_the_network_configuration(aws_credentials):
    session = aws_credentials.get_boto3_session()
    ec2_resource = session.resource("ec2")
    vpc = ec2_resource.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2_resource.create_subnet(CidrBlock="10.0.2.0/24", VpcId=vpc.id)

    task = ECSTask(
        aws_credentials=aws_credentials,
        task_customizations=[
            {
                "op": "add",
                "path": "/networkConfiguration/awsvpcConfiguration/subnets",
                "value": [subnet.id],
            },
            {
                "op": "add",
                "path": "/networkConfiguration/awsvpcConfiguration/assignPublicIp",
                "value": "DISABLED",
            },
        ],
    )

    # Capture the task run call because moto does not track 'networkConfiguration'
    original_run_task = task._run_task
    mock_run_task = MagicMock(side_effect=original_run_task)
    task._run_task = mock_run_task

    print(task.preview())

    await run_then_stop_task(task)

    network_configuration = mock_run_task.call_args[0][1].get("networkConfiguration")

    # Subnet ids are copied from the vpc
    assert network_configuration == {
        "awsvpcConfiguration": {
            "subnets": [subnet.id],
            "assignPublicIp": "DISABLED",
            "securityGroups": [],
        }
    }


@pytest.mark.usefixtures("ecs_mocks")
async def test_task_customizations_as_string(aws_credentials):
    tc = (
        '[{"op": "replace", "path": "/overrides/cpu", "value": "512"}, '
        '{"op": "replace", "path": "/overrides/memory", "value": "1024"}]'
    )

    task = ECSTask(
        aws_credentials=aws_credentials, memory=512, cpu=256, task_customizations=tc
    )  # type: ignore

    original_run_task = task._run_task
    mock_run_task = MagicMock(side_effect=original_run_task)
    task._run_task = mock_run_task

    await run_then_stop_task(task)

    overrides = mock_run_task.call_args[0][1].get("overrides")

    assert overrides["memory"] == "1024"
    assert overrides["cpu"] == "512"


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize(
    "fields,prepare_inputs,expected_family",
    [
        # Default
        ({}, {}, "prefect"),
        # Only flow
        ({}, {"flow": Flow(name="foo")}, "prefect__foo"),
        # Only deployment
        (
            {},
            {"deployment": Deployment.construct(name="foo")},
            "prefect__unknown-flow__foo",
        ),
        # Flow and deployment
        (
            {},
            {
                "flow": Flow(name="foo"),
                "deployment": Deployment.construct(name="bar"),
            },
            "prefect__foo__bar",
        ),
        # Family provided as a field
        (
            {"family": "test"},
            {
                "flow": Flow(name="foo"),
                "deployment": Deployment.construct(name="bar"),
            },
            "test",
        ),
        # Family provided in a task definition
        (
            {"task_definition": {"family": "test"}},
            {
                "flow": Flow(name="foo"),
                "deployment": Deployment.construct(name="bar"),
            },
            "test",
        ),
    ],
)
async def test_family_from_flow_run_metadata(
    aws_credentials, fields, prepare_inputs, expected_family
):
    prepare_inputs.setdefault("flow_run", FlowRun.construct())

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        **fields,
    ).prepare_for_flow_run(**prepare_inputs)
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    assert task_definition["family"] == expected_family


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize(
    "given_family,expected_family",
    [
        # Default
        (None, "prefect"),
        ("", "prefect"),
        # Length limited to 255
        ("x" * 300, "x" * 255),
        # Spaces are not allowed
        ("foo bar", "foo-bar"),
        # Special characters are not allowed
        ("foo*bar&!", "foo-bar"),
    ],
)
async def test_user_provided_family(aws_credentials, given_family, expected_family):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        family=given_family,
    )
    print(task.preview())

    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    assert task_definition["family"] == expected_family


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize("prepare_for_flow_run", [True, False])
async def test_family_from_task_definition_arn(aws_credentials, prepare_for_flow_run):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task_definition_arn = ecs_client.register_task_definition(
        **{**BASE_TASK_DEFINITION, "family": "test-family"}
    )["taskDefinition"]["taskDefinitionArn"]

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        task_definition_arn=task_definition_arn,
        launch_type="EC2",
        image=None,
    )
    if prepare_for_flow_run:
        task = task.prepare_for_flow_run(
            flow_run=FlowRun.construct(),
            flow=Flow(name="foo"),
            deployment=Deployment.construct(name="bar"),
        )

    print(task.preview())

    task_arn = await run_then_stop_task(task)

    task = describe_task(ecs_client, task_arn)
    task_definition = describe_task_definition(ecs_client, task)
    assert task_definition["family"] == "test-family"


@pytest.mark.usefixtures("ecs_mocks")
@pytest.mark.parametrize(
    "cluster", [None, "default", "second-cluster", "second-cluster-arn"]
)
async def test_kill(aws_credentials, cluster: str):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    # Kill requires cluster-specificity so we test with variable clusters
    second_cluster_arn = create_test_ecs_cluster(ecs_client, "second-cluster")
    add_ec2_instance_to_ecs_cluster(session, "second-cluster")

    if cluster == "second-cluster-arn":
        # Use the actual arn for this test case
        cluster = second_cluster_arn

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        cluster=cluster,
    )
    print(task.preview())

    with anyio.fail_after(5):
        async with anyio.create_task_group() as tg:
            identifier = await tg.start(task.run)

            await task.kill(identifier)

    _, task_arn = parse_task_identifier(identifier)
    task = describe_task(ecs_client, task_arn)
    assert task["lastStatus"] == "STOPPED"


@pytest.mark.usefixtures("ecs_mocks")
async def test_kill_with_invalid_identifier(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["sleep", "1000"],
    )
    print(task.preview())

    with pytest.raises(ValueError):
        await task.kill("test")


@pytest.mark.usefixtures("ecs_mocks")
async def test_kill_with_mismatched_cluster(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["sleep", "1000"],
        cluster="foo",
    )
    print(task.preview())

    with pytest.raises(
        InfrastructureNotAvailable,
        match=(
            "Cannot stop ECS task: this infrastructure block has access to cluster "
            "'foo' but the task is running in cluster 'bar'."
        ),
    ):
        await task.kill("bar:::task_arn")


@pytest.mark.usefixtures("ecs_mocks")
async def test_kill_with_cluster_that_does_not_exist(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["sleep", "1000"],
        cluster="foo",
    )
    print(task.preview())

    with pytest.raises(
        InfrastructureNotFound,
        match="Cannot stop ECS task: the cluster 'foo' could not be found.",
    ):
        await task.kill("foo::task_arn")


@pytest.mark.usefixtures("ecs_mocks")
async def test_kill_with_task_that_does_not_exist(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["sleep", "1000"],
        cluster="default",
    )
    print(task.preview())

    # Run the task so that a task definition is registered in the cluster
    await run_then_stop_task(task)

    with pytest.raises(
        InfrastructureNotFound,
        match=(
            "Cannot stop ECS task: the task 'foo' could not be found in cluster"
            " 'default'"
        ),
    ):
        await task.kill("default::foo")


@pytest.mark.usefixtures("ecs_mocks")
async def test_kill_with_cluster_that_has_no_tasks(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["sleep", "1000"],
        cluster="default",
    )
    print(task.preview())

    with pytest.raises(
        InfrastructureNotFound,
        match="Cannot stop ECS task: the cluster 'default' has no tasks.",
    ):
        await task.kill("default::foo")


@pytest.mark.usefixtures("ecs_mocks")
async def test_kill_with_task_that_is_already_stopped(aws_credentials):
    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
        command=["sleep", "1000"],
        cluster="default",
    )
    print(task.preview())

    # Run and stop the task
    task_arn = await run_then_stop_task(task)

    # AWS will happily stop the task "again"
    await task.kill(f"default::{task_arn}")


@pytest.mark.usefixtures("ecs_mocks")
async def test_kill_with_grace_period(aws_credentials, caplog):
    session = aws_credentials.get_boto3_session()
    ecs_client = session.client("ecs")

    task = ECSTask(
        aws_credentials=aws_credentials,
        auto_deregister_task_definition=False,
    )
    print(task.preview())

    with anyio.fail_after(5):
        async with anyio.create_task_group() as tg:
            identifier = await tg.start(task.run)

            await task.kill(identifier, grace_seconds=60)

    # Task stops correctly
    _, task_arn = parse_task_identifier(identifier)
    task = describe_task(ecs_client, task_arn)
    assert task["lastStatus"] == "STOPPED"

    # Logs warning
    assert "grace period of 60s requested, but AWS does not support" in caplog.text



"""Tasks for interacting with AWS S3"""
import asyncio
import io
import os
import uuid
from pathlib import Path
from typing import Any, BinaryIO, Dict, List, Optional, Union

import boto3
from botocore.paginate import PageIterator
from botocore.response import StreamingBody
from prefect import get_run_logger, task
from prefect.blocks.abstract import ObjectStorageBlock
from prefect.filesystems import WritableDeploymentStorage, WritableFileSystem
from prefect.utilities.asyncutils import run_sync_in_worker_thread, sync_compatible
from prefect.utilities.filesystem import filter_files
from pydantic import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    from pydantic.v1 import Field
else:
    from pydantic import Field

from prefect_aws import AwsCredentials, MinIOCredentials
from prefect_aws.client_parameters import AwsClientParameters


@task
async def s3_download(
    bucket: str,
    key: str,
    aws_credentials: AwsCredentials,
    aws_client_parameters: AwsClientParameters = AwsClientParameters(),
) -> bytes:
    """
    Downloads an object with a given key from a given S3 bucket.

    Args:
        bucket: Name of bucket to download object from. Required if a default value was
            not supplied when creating the task.
        key: Key of object to download. Required if a default value was not supplied
            when creating the task.
        aws_credentials: Credentials to use for authentication with AWS.
        aws_client_parameters: Custom parameter for the boto3 client initialization.


    Returns:
        A `bytes` representation of the downloaded object.

    Example:
        Download a file from an S3 bucket:

        ```python
        from prefect import flow
        from prefect_aws import AwsCredentials
        from prefect_aws.s3 import s3_download


        @flow
        async def example_s3_download_flow():
            aws_credentials = AwsCredentials(
                aws_access_key_id="acccess_key_id",
                aws_secret_access_key="secret_access_key"
            )
            data = await s3_download(
                bucket="bucket",
                key="key",
                aws_credentials=aws_credentials,
            )

        example_s3_download_flow()
        ```
    """
    logger = get_run_logger()
    logger.info("Downloading object from bucket %s with key %s", bucket, key)

    s3_client = aws_credentials.get_boto3_session().client(
        "s3", **aws_client_parameters.get_params_override()
    )
    stream = io.BytesIO()
    await run_sync_in_worker_thread(
        s3_client.download_fileobj, Bucket=bucket, Key=key, Fileobj=stream
    )
    stream.seek(0)
    output = stream.read()

    return output


@task
async def s3_upload(
    data: bytes,
    bucket: str,
    aws_credentials: AwsCredentials,
    aws_client_parameters: AwsClientParameters = AwsClientParameters(),
    key: Optional[str] = None,
) -> str:
    """
    Uploads data to an S3 bucket.

    Args:
        data: Bytes representation of data to upload to S3.
        bucket: Name of bucket to upload data to. Required if a default value was not
            supplied when creating the task.
        aws_credentials: Credentials to use for authentication with AWS.
        aws_client_parameters: Custom parameter for the boto3 client initialization..
        key: Key of object to download. Defaults to a UUID string.

    Returns:
        The key of the uploaded object

    Example:
        Read and upload a file to an S3 bucket:

        ```python
        from prefect import flow
        from prefect_aws import AwsCredentials
        from prefect_aws.s3 import s3_upload


        @flow
        async def example_s3_upload_flow():
            aws_credentials = AwsCredentials(
                aws_access_key_id="acccess_key_id",
                aws_secret_access_key="secret_access_key"
            )
            with open("data.csv", "rb") as file:
                key = await s3_upload(
                    bucket="bucket",
                    key="data.csv",
                    data=file.read(),
                    aws_credentials=aws_credentials,
                )

        example_s3_upload_flow()
        ```
    """
    logger = get_run_logger()

    key = key or str(uuid.uuid4())

    logger.info("Uploading object to bucket %s with key %s", bucket, key)

    s3_client = aws_credentials.get_boto3_session().client(
        "s3", **aws_client_parameters.get_params_override()
    )
    stream = io.BytesIO(data)
    await run_sync_in_worker_thread(
        s3_client.upload_fileobj, stream, Bucket=bucket, Key=key
    )

    return key


@task
async def s3_copy(
    source_path: str,
    target_path: str,
    source_bucket_name: str,
    aws_credentials: AwsCredentials,
    target_bucket_name: Optional[str] = None,
    **copy_kwargs,
) -> str:
    """Uses S3's internal
    [CopyObject](https://docs.aws.amazon.com/AmazonS3/latest/API/API_CopyObject.html)
    to copy objects within or between buckets. To copy objects between buckets, the
    credentials must have permission to read the source object and write to the target
    object. If the credentials do not have those permissions, try using
    `S3Bucket.stream_from`.

    Args:
        source_path: The path to the object to copy. Can be a string or `Path`.
        target_path: The path to copy the object to. Can be a string or `Path`.
        source_bucket_name: The bucket to copy the object from.
        aws_credentials: Credentials to use for authentication with AWS.
        target_bucket_name: The bucket to copy the object to. If not provided, defaults
            to `source_bucket`.
        **copy_kwargs: Additional keyword arguments to pass to `S3Client.copy_object`.

    Returns:
        The path that the object was copied to. Excludes the bucket name.

    Examples:

        Copy notes.txt from s3://my-bucket/my_folder/notes.txt to
        s3://my-bucket/my_folder/notes_copy.txt.

        ```python
        from prefect import flow
        from prefect_aws import AwsCredentials
        from prefect_aws.s3 import s3_copy

        aws_credentials = AwsCredentials.load("my-creds")

        @flow
        async def example_copy_flow():
            await s3_copy(
                source_path="my_folder/notes.txt",
                target_path="my_folder/notes_copy.txt",
                source_bucket_name="my-bucket",
                aws_credentials=aws_credentials,
            )

        example_copy_flow()
        ```

        Copy notes.txt from s3://my-bucket/my_folder/notes.txt to
        s3://other-bucket/notes_copy.txt.

        ```python
        from prefect import flow
        from prefect_aws import AwsCredentials
        from prefect_aws.s3 import s3_copy

        aws_credentials = AwsCredentials.load("shared-creds")

        @flow
        async def example_copy_flow():
            await s3_copy(
                source_path="my_folder/notes.txt",
                target_path="notes_copy.txt",
                source_bucket_name="my-bucket",
                aws_credentials=aws_credentials,
                target_bucket_name="other-bucket",
            )

        example_copy_flow()
        ```

    """
    logger = get_run_logger()

    s3_client = aws_credentials.get_s3_client()

    target_bucket_name = target_bucket_name or source_bucket_name

    logger.info(
        "Copying object from bucket %s with key %s to bucket %s with key %s",
        source_bucket_name,
        source_path,
        target_bucket_name,
        target_path,
    )

    s3_client.copy_object(
        CopySource={"Bucket": source_bucket_name, "Key": source_path},
        Bucket=target_bucket_name,
        Key=target_path,
        **copy_kwargs,
    )

    return target_path


@task
async def s3_move(
    source_path: str,
    target_path: str,
    source_bucket_name: str,
    aws_credentials: AwsCredentials,
    target_bucket_name: Optional[str] = None,
) -> str:
    """
    Move an object from one S3 location to another. To move objects between buckets,
    the credentials must have permission to read and delete the source object and write
    to the target object. If the credentials do not have those permissions, this method
    will raise an error. If the credentials have permission to read the source object
    but not delete it, the object will be copied but not deleted.

    Args:
        source_path: The path of the object to move
        target_path: The path to move the object to
        source_bucket_name: The name of the bucket containing the source object
        aws_credentials: Credentials to use for authentication with AWS.
        target_bucket_name: The bucket to copy the object to. If not provided, defaults
            to `source_bucket`.

    Returns:
        The path that the object was moved to. Excludes the bucket name.
    """
    logger = get_run_logger()

    s3_client = aws_credentials.get_s3_client()

    # If target bucket is not provided, assume it's the same as the source bucket
    target_bucket_name = target_bucket_name or source_bucket_name

    logger.info(
        "Moving object from s3://%s/%s s3://%s/%s",
        source_bucket_name,
        source_path,
        target_bucket_name,
        target_path,
    )

    # Copy the object to the new location
    s3_client.copy_object(
        Bucket=target_bucket_name,
        CopySource={"Bucket": source_bucket_name, "Key": source_path},
        Key=target_path,
    )

    # Delete the original object
    s3_client.delete_object(Bucket=source_bucket_name, Key=source_path)

    return target_path


def _list_objects_sync(page_iterator: PageIterator):
    """
    Synchronous method to collect S3 objects into a list

    Args:
        page_iterator: AWS Paginator for S3 objects

    Returns:
        List[Dict]: List of object information
    """
    return [content for page in page_iterator for content in page.get("Contents", [])]


@task
async def s3_list_objects(
    bucket: str,
    aws_credentials: AwsCredentials,
    aws_client_parameters: AwsClientParameters = AwsClientParameters(),
    prefix: str = "",
    delimiter: str = "",
    page_size: Optional[int] = None,
    max_items: Optional[int] = None,
    jmespath_query: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    Lists details of objects in a given S3 bucket.

    Args:
        bucket: Name of bucket to list items from. Required if a default value was not
            supplied when creating the task.
        aws_credentials: Credentials to use for authentication with AWS.
        aws_client_parameters: Custom parameter for the boto3 client initialization..
        prefix: Used to filter objects with keys starting with the specified prefix.
        delimiter: Character used to group keys of listed objects.
        page_size: Number of objects to return in each request to the AWS API.
        max_items: Maximum number of objects that to be returned by task.
        jmespath_query: Query used to filter objects based on object attributes refer to
            the [boto3 docs](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/paginators.html#filtering-results-with-jmespath)
            for more information on how to construct queries.

    Returns:
        A list of dictionaries containing information about the objects retrieved. Refer
            to the boto3 docs for an example response.

    Example:
        List all objects in a bucket:

        ```python
        from prefect import flow
        from prefect_aws import AwsCredentials
        from prefect_aws.s3 import s3_list_objects


        @flow
        async def example_s3_list_objects_flow():
            aws_credentials = AwsCredentials(
                aws_access_key_id="acccess_key_id",
                aws_secret_access_key="secret_access_key"
            )
            objects = await s3_list_objects(
                bucket="data_bucket",
                aws_credentials=aws_credentials
            )

        example_s3_list_objects_flow()
        ```
    """  # noqa E501
    logger = get_run_logger()
    logger.info("Listing objects in bucket %s with prefix %s", bucket, prefix)

    s3_client = aws_credentials.get_boto3_session().client(
        "s3", **aws_client_parameters.get_params_override()
    )
    paginator = s3_client.get_paginator("list_objects_v2")
    page_iterator = paginator.paginate(
        Bucket=bucket,
        Prefix=prefix,
        Delimiter=delimiter,
        PaginationConfig={"PageSize": page_size, "MaxItems": max_items},
    )
    if jmespath_query:
        page_iterator = page_iterator.search(f"{jmespath_query} | {{Contents: @}}")

    return await run_sync_in_worker_thread(_list_objects_sync, page_iterator)


class S3Bucket(WritableFileSystem, WritableDeploymentStorage, ObjectStorageBlock):

    """
    Block used to store data using AWS S3 or S3-compatible object storage like MinIO.

    Attributes:
        bucket_name: Name of your bucket.
        credentials: A block containing your credentials to AWS or MinIO.
        bucket_folder: A default path to a folder within the S3 bucket to use
            for reading and writing objects.
    """

    _logo_url = "https://cdn.sanity.io/images/3ugk85nk/production/d74b16fe84ce626345adf235a47008fea2869a60-225x225.png"  # noqa
    _block_type_name = "S3 Bucket"
    _documentation_url = (
        "https://prefecthq.github.io/prefect-aws/s3/#prefect_aws.s3.S3Bucket"  # noqa
    )

    bucket_name: str = Field(default=..., description="Name of your bucket.")

    credentials: Union[AwsCredentials, MinIOCredentials] = Field(
        default_factory=AwsCredentials,
        description="A block containing your credentials to AWS or MinIO.",
    )

    bucket_folder: str = Field(
        default="",
        description=(
            "A default path to a folder within the S3 bucket to use "
            "for reading and writing objects."
        ),
    )

    class Config:
        smart_union = True

    # Property to maintain compatibility with storage block based deployments
    @property
    def basepath(self) -> str:
        """
        The base path of the S3 bucket.

        Returns:
            str: The base path of the S3 bucket.
        """
        return self.bucket_folder

    @basepath.setter
    def basepath(self, value: str) -> None:
        self.bucket_folder = value

    def _resolve_path(self, path: str) -> str:
        """
        A helper function used in write_path to join `self.basepath` and `path`.

        Args:

            path: Name of the key, e.g. "file1". Each object in your
                bucket has a unique key (or key name).

        """
        # If bucket_folder provided, it means we won't write to the root dir of
        # the bucket. So we need to add it on the front of the path.
        #
        # AWS object key naming guidelines require '/' for bucket folders.
        # Get POSIX path to prevent `pathlib` from inferring '\' on Windows OS
        path = (
            (Path(self.bucket_folder) / path).as_posix() if self.bucket_folder else path
        )

        return path

    def _get_s3_client(self) -> boto3.client:
        """
        Authenticate MinIO credentials or AWS credentials and return an S3 client.
        This is a helper function called by read_path() or write_path().
        """
        return self.credentials.get_s3_client()

    def _get_bucket_resource(self) -> boto3.resource:
        """
        Retrieves boto3 resource object for the configured bucket
        """
        params_override = self.credentials.aws_client_parameters.get_params_override()
        bucket = (
            self.credentials.get_boto3_session()
            .resource("s3", **params_override)
            .Bucket(self.bucket_name)
        )
        return bucket

    @sync_compatible
    async def get_directory(
        self, from_path: Optional[str] = None, local_path: Optional[str] = None
    ) -> None:
        """
        Copies a folder from the configured S3 bucket to a local directory.

        Defaults to copying the entire contents of the block's basepath to the current
        working directory.

        Args:
            from_path: Path in S3 bucket to download from. Defaults to the block's
                configured basepath.
            local_path: Local path to download S3 contents to. Defaults to the current
                working directory.
        """
        bucket_folder = self.bucket_folder
        if from_path is None:
            from_path = str(bucket_folder) if bucket_folder else ""

        if local_path is None:
            local_path = str(Path(".").absolute())
        else:
            local_path = str(Path(local_path).expanduser())

        bucket = self._get_bucket_resource()
        for obj in bucket.objects.filter(Prefix=from_path):
            if obj.key[-1] == "/":
                # object is a folder and will be created if it contains any objects
                continue
            target = os.path.join(
                local_path,
                os.path.relpath(obj.key, from_path),
            )
            os.makedirs(os.path.dirname(target), exist_ok=True)
            bucket.download_file(obj.key, target)

    @sync_compatible
    async def put_directory(
        self,
        local_path: Optional[str] = None,
        to_path: Optional[str] = None,
        ignore_file: Optional[str] = None,
    ) -> int:
        """
        Uploads a directory from a given local path to the configured S3 bucket in a
        given folder.

        Defaults to uploading the entire contents the current working directory to the
        block's basepath.

        Args:
            local_path: Path to local directory to upload from.
            to_path: Path in S3 bucket to upload to. Defaults to block's configured
                basepath.
            ignore_file: Path to file containing gitignore style expressions for
                filepaths to ignore.

        """
        to_path = "" if to_path is None else to_path

        if local_path is None:
            local_path = "."

        included_files = None
        if ignore_file:
            with open(ignore_file, "r") as f:
                ignore_patterns = f.readlines()

            included_files = filter_files(local_path, ignore_patterns)

        uploaded_file_count = 0
        for local_file_path in Path(local_path).expanduser().rglob("*"):
            if (
                included_files is not None
                and str(local_file_path.relative_to(local_path)) not in included_files
            ):
                continue
            elif not local_file_path.is_dir():
                remote_file_path = Path(to_path) / local_file_path.relative_to(
                    local_path
                )
                with open(local_file_path, "rb") as local_file:
                    local_file_content = local_file.read()

                await self.write_path(
                    remote_file_path.as_posix(), content=local_file_content
                )
                uploaded_file_count += 1

        return uploaded_file_count

    @sync_compatible
    async def read_path(self, path: str) -> bytes:
        """
        Read specified path from S3 and return contents. Provide the entire
        path to the key in S3.

        Args:
            path: Entire path to (and including) the key.

        Example:
            Read "subfolder/file1" contents from an S3 bucket named "bucket":
            ```python
            from prefect_aws import AwsCredentials
            from prefect_aws.s3 import S3Bucket

            aws_creds = AwsCredentials(
                aws_access_key_id=AWS_ACCESS_KEY_ID,
                aws_secret_access_key=AWS_SECRET_ACCESS_KEY
            )

            s3_bucket_block = S3Bucket(
                bucket_name="bucket",
                aws_credentials=aws_creds,
                basepath="subfolder"
            )

            key_contents = s3_bucket_block.read_path(path="subfolder/file1")
            ```
        """
        path = self._resolve_path(path)

        return await run_sync_in_worker_thread(self._read_sync, path)

    def _read_sync(self, key: str) -> bytes:
        """
        Called by read_path(). Creates an S3 client and retrieves the
        contents from  a specified path.
        """

        s3_client = self._get_s3_client()

        with io.BytesIO() as stream:
            s3_client.download_fileobj(Bucket=self.bucket_name, Key=key, Fileobj=stream)
            stream.seek(0)
            output = stream.read()
            return output

    @sync_compatible
    async def write_path(self, path: str, content: bytes) -> str:
        """
        Writes to an S3 bucket.

        Args:

            path: The key name. Each object in your bucket has a unique
                key (or key name).
            content: What you are uploading to S3.

        Example:

            Write data to the path `dogs/small_dogs/havanese` in an S3 Bucket:
            ```python
            from prefect_aws import MinioCredentials
            from prefect_aws.s3 import S3Bucket

            minio_creds = MinIOCredentials(
                minio_root_user = "minioadmin",
                minio_root_password = "minioadmin",
            )

            s3_bucket_block = S3Bucket(
                bucket_name="bucket",
                minio_credentials=minio_creds,
                basepath="dogs/smalldogs",
                endpoint_url="http://localhost:9000",
            )
            s3_havanese_path = s3_bucket_block.write_path(path="havanese", content=data)
            ```
        """

        path = self._resolve_path(path)

        await run_sync_in_worker_thread(self._write_sync, path, content)

        return path

    def _write_sync(self, key: str, data: bytes) -> None:
        """
        Called by write_path(). Creates an S3 client and uploads a file
        object.
        """

        s3_client = self._get_s3_client()

        with io.BytesIO(data) as stream:
            s3_client.upload_fileobj(Fileobj=stream, Bucket=self.bucket_name, Key=key)

    # NEW BLOCK INTERFACE METHODS BELOW
    @staticmethod
    def _list_objects_sync(page_iterator: PageIterator) -> List[Dict[str, Any]]:
        """
        Synchronous method to collect S3 objects into a list

        Args:
            page_iterator: AWS Paginator for S3 objects

        Returns:
            List[Dict]: List of object information
        """
        return [
            content for page in page_iterator for content in page.get("Contents", [])
        ]

    def _join_bucket_folder(self, bucket_path: str = "") -> str:
        """
        Joins the base bucket folder to the bucket path.
        NOTE: If a method reuses another method in this class, be careful to not
        call this  twice because it'll join the bucket folder twice.
        See https://github.com/PrefectHQ/prefect-aws/issues/141 for a past issue.
        """
        if not self.bucket_folder and not bucket_path:
            # there's a difference between "." and "", at least in the tests
            return ""

        bucket_path = str(bucket_path)
        if self.bucket_folder != "" and bucket_path.startswith(self.bucket_folder):
            self.logger.info(
                f"Bucket path {bucket_path!r} is already prefixed with "
                f"bucket folder {self.bucket_folder!r}; is this intentional?"
            )

        return (Path(self.bucket_folder) / bucket_path).as_posix() + (
            "" if not bucket_path.endswith("/") else "/"
        )

    @sync_compatible
    async def list_objects(
        self,
        folder: str = "",
        delimiter: str = "",
        page_size: Optional[int] = None,
        max_items: Optional[int] = None,
        jmespath_query: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        Args:
            folder: Folder to list objects from.
            delimiter: Character used to group keys of listed objects.
            page_size: Number of objects to return in each request to the AWS API.
            max_items: Maximum number of objects that to be returned by task.
            jmespath_query: Query used to filter objects based on object attributes refer to
                the [boto3 docs](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/paginators.html#filtering-results-with-jmespath)
                for more information on how to construct queries.

        Returns:
            List of objects and their metadata in the bucket.

        Examples:
            List objects under the `base_folder`.
            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            s3_bucket.list_objects("base_folder")
            ```
        """  # noqa: E501
        bucket_path = self._join_bucket_folder(folder)
        client = self.credentials.get_s3_client()
        paginator = client.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(
            Bucket=self.bucket_name,
            Prefix=bucket_path,
            Delimiter=delimiter,
            PaginationConfig={"PageSize": page_size, "MaxItems": max_items},
        )
        if jmespath_query:
            page_iterator = page_iterator.search(f"{jmespath_query} | {{Contents: @}}")

        self.logger.info(f"Listing objects in bucket {bucket_path}.")
        objects = await run_sync_in_worker_thread(
            self._list_objects_sync, page_iterator
        )
        return objects

    @sync_compatible
    async def download_object_to_path(
        self,
        from_path: str,
        to_path: Optional[Union[str, Path]],
        **download_kwargs: Dict[str, Any],
    ) -> Path:
        """
        Downloads an object from the S3 bucket to a path.

        Args:
            from_path: The path to the object to download; this gets prefixed
                with the bucket_folder.
            to_path: The path to download the object to. If not provided, the
                object's name will be used.
            **download_kwargs: Additional keyword arguments to pass to
                `Client.download_file`.

        Returns:
            The absolute path that the object was downloaded to.

        Examples:
            Download my_folder/notes.txt object to notes.txt.
            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            s3_bucket.download_object_to_path("my_folder/notes.txt", "notes.txt")
            ```
        """
        if to_path is None:
            to_path = Path(from_path).name

        # making path absolute, but converting back to str here
        # since !r looks nicer that way and filename arg expects str
        to_path = str(Path(to_path).absolute())
        bucket_path = self._join_bucket_folder(from_path)
        client = self.credentials.get_s3_client()

        self.logger.debug(
            f"Preparing to download object from bucket {self.bucket_name!r} "
            f"path {bucket_path!r} to {to_path!r}."
        )
        await run_sync_in_worker_thread(
            client.download_file,
            Bucket=self.bucket_name,
            Key=from_path,
            Filename=to_path,
            **download_kwargs,
        )
        self.logger.info(
            f"Downloaded object from bucket {self.bucket_name!r} path {bucket_path!r} "
            f"to {to_path!r}."
        )
        return Path(to_path)

    @sync_compatible
    async def download_object_to_file_object(
        self,
        from_path: str,
        to_file_object: BinaryIO,
        **download_kwargs: Dict[str, Any],
    ) -> BinaryIO:
        """
        Downloads an object from the object storage service to a file-like object,
        which can be a BytesIO object or a BufferedWriter.

        Args:
            from_path: The path to the object to download from; this gets prefixed
                with the bucket_folder.
            to_file_object: The file-like object to download the object to.
            **download_kwargs: Additional keyword arguments to pass to
                `Client.download_fileobj`.

        Returns:
            The file-like object that the object was downloaded to.

        Examples:
            Download my_folder/notes.txt object to a BytesIO object.
            ```python
            from io import BytesIO

            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            with BytesIO() as buf:
                s3_bucket.download_object_to_file_object("my_folder/notes.txt", buf)
            ```

            Download my_folder/notes.txt object to a BufferedWriter.
            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            with open("notes.txt", "wb") as f:
                s3_bucket.download_object_to_file_object("my_folder/notes.txt", f)
            ```
        """
        client = self.credentials.get_s3_client()
        bucket_path = self._join_bucket_folder(from_path)

        self.logger.debug(
            f"Preparing to download object from bucket {self.bucket_name!r} "
            f"path {bucket_path!r} to file object."
        )
        await run_sync_in_worker_thread(
            client.download_fileobj,
            Bucket=self.bucket_name,
            Key=bucket_path,
            Fileobj=to_file_object,
            **download_kwargs,
        )
        self.logger.info(
            f"Downloaded object from bucket {self.bucket_name!r} path {bucket_path!r} "
            "to file object."
        )
        return to_file_object

    @sync_compatible
    async def download_folder_to_path(
        self,
        from_folder: str,
        to_folder: Optional[Union[str, Path]] = None,
        **download_kwargs: Dict[str, Any],
    ) -> Path:
        """
        Downloads objects *within* a folder (excluding the folder itself)
        from the S3 bucket to a folder.

        Args:
            from_folder: The path to the folder to download from.
            to_folder: The path to download the folder to.
            **download_kwargs: Additional keyword arguments to pass to
                `Client.download_file`.

        Returns:
            The absolute path that the folder was downloaded to.

        Examples:
            Download my_folder to a local folder named my_folder.
            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            s3_bucket.download_folder_to_path("my_folder", "my_folder")
            ```
        """
        if to_folder is None:
            to_folder = ""
        to_folder = Path(to_folder).absolute()

        client = self.credentials.get_s3_client()
        objects = await self.list_objects(folder=from_folder)

        # do not call self._join_bucket_folder for filter
        # because it's built-in to that method already!
        # however, we still need to do it because we're using relative_to
        bucket_folder = self._join_bucket_folder(from_folder)

        async_coros = []
        for object in objects:
            bucket_path = Path(object["Key"]).relative_to(bucket_folder)
            # this skips the actual directory itself, e.g.
            # `my_folder/` will be skipped
            # `my_folder/notes.txt` will be downloaded
            if bucket_path.is_dir():
                continue
            to_path = to_folder / bucket_path
            to_path.parent.mkdir(parents=True, exist_ok=True)
            to_path = str(to_path)  # must be string
            self.logger.info(
                f"Downloading object from bucket {self.bucket_name!r} path "
                f"{bucket_path.as_posix()!r} to {to_path!r}."
            )
            async_coros.append(
                run_sync_in_worker_thread(
                    client.download_file,
                    Bucket=self.bucket_name,
                    Key=object["Key"],
                    Filename=to_path,
                    **download_kwargs,
                )
            )
        await asyncio.gather(*async_coros)

        return Path(to_folder)

    @sync_compatible
    async def stream_from(
        self,
        bucket: "S3Bucket",
        from_path: str,
        to_path: Optional[str] = None,
        **upload_kwargs: Dict[str, Any],
    ) -> str:
        """Streams an object from another bucket to this bucket. Requires the
        object to be downloaded and uploaded in chunks. If `self`'s credentials
        allow for writes to the other bucket, try using `S3Bucket.copy_object`.

        Args:
            bucket: The bucket to stream from.
            from_path: The path of the object to stream.
            to_path: The path to stream the object to. Defaults to the object's name.
            **upload_kwargs: Additional keyword arguments to pass to
                `Client.upload_fileobj`.

        Returns:
            The path that the object was uploaded to.

        Examples:
            Stream notes.txt from your-bucket/notes.txt to my-bucket/landed/notes.txt.

            ```python
            from prefect_aws.s3 import S3Bucket

            your_s3_bucket = S3Bucket.load("your-bucket")
            my_s3_bucket = S3Bucket.load("my-bucket")

            my_s3_bucket.stream_from(
                your_s3_bucket,
                "notes.txt",
                to_path="landed/notes.txt"
            )
            ```

        """
        if to_path is None:
            to_path = Path(from_path).name

        # Get the source object's StreamingBody
        from_path: str = bucket._join_bucket_folder(from_path)
        from_client = bucket.credentials.get_s3_client()
        obj = await run_sync_in_worker_thread(
            from_client.get_object, Bucket=bucket.bucket_name, Key=from_path
        )
        body: StreamingBody = obj["Body"]

        # Upload the StreamingBody to this bucket
        bucket_path = str(self._join_bucket_folder(to_path))
        to_client = self.credentials.get_s3_client()
        await run_sync_in_worker_thread(
            to_client.upload_fileobj,
            Fileobj=body,
            Bucket=self.bucket_name,
            Key=bucket_path,
            **upload_kwargs,
        )
        self.logger.info(
            f"Streamed s3://{bucket.bucket_name}/{from_path} to the bucket "
            f"{self.bucket_name!r} path {bucket_path!r}."
        )
        return bucket_path

    @sync_compatible
    async def upload_from_path(
        self,
        from_path: Union[str, Path],
        to_path: Optional[str] = None,
        **upload_kwargs: Dict[str, Any],
    ) -> str:
        """
        Uploads an object from a path to the S3 bucket.

        Args:
            from_path: The path to the file to upload from.
            to_path: The path to upload the file to.
            **upload_kwargs: Additional keyword arguments to pass to `Client.upload`.

        Returns:
            The path that the object was uploaded to.

        Examples:
            Upload notes.txt to my_folder/notes.txt.
            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            s3_bucket.upload_from_path("notes.txt", "my_folder/notes.txt")
            ```
        """
        from_path = str(Path(from_path).absolute())
        if to_path is None:
            to_path = Path(from_path).name

        bucket_path = str(self._join_bucket_folder(to_path))
        client = self.credentials.get_s3_client()

        await run_sync_in_worker_thread(
            client.upload_file,
            Filename=from_path,
            Bucket=self.bucket_name,
            Key=bucket_path,
            **upload_kwargs,
        )
        self.logger.info(
            f"Uploaded from {from_path!r} to the bucket "
            f"{self.bucket_name!r} path {bucket_path!r}."
        )
        return bucket_path

    @sync_compatible
    async def upload_from_file_object(
        self, from_file_object: BinaryIO, to_path: str, **upload_kwargs: Dict[str, Any]
    ) -> str:
        """
        Uploads an object to the S3 bucket from a file-like object,
        which can be a BytesIO object or a BufferedReader.

        Args:
            from_file_object: The file-like object to upload from.
            to_path: The path to upload the object to.
            **upload_kwargs: Additional keyword arguments to pass to
                `Client.upload_fileobj`.

        Returns:
            The path that the object was uploaded to.

        Examples:
            Upload BytesIO object to my_folder/notes.txt.
            ```python
            from io import BytesIO

            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            with open("notes.txt", "rb") as f:
                s3_bucket.upload_from_file_object(f, "my_folder/notes.txt")
            ```

            Upload BufferedReader object to my_folder/notes.txt.
            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            with open("notes.txt", "rb") as f:
                s3_bucket.upload_from_file_object(
                    f, "my_folder/notes.txt"
                )
            ```
        """
        bucket_path = str(self._join_bucket_folder(to_path))
        client = self.credentials.get_s3_client()
        await run_sync_in_worker_thread(
            client.upload_fileobj,
            Fileobj=from_file_object,
            Bucket=self.bucket_name,
            Key=bucket_path,
            **upload_kwargs,
        )
        self.logger.info(
            "Uploaded from file object to the bucket "
            f"{self.bucket_name!r} path {bucket_path!r}."
        )
        return bucket_path

    @sync_compatible
    async def upload_from_folder(
        self,
        from_folder: Union[str, Path],
        to_folder: Optional[str] = None,
        **upload_kwargs: Dict[str, Any],
    ) -> str:
        """
        Uploads files *within* a folder (excluding the folder itself)
        to the object storage service folder.

        Args:
            from_folder: The path to the folder to upload from.
            to_folder: The path to upload the folder to.
            **upload_kwargs: Additional keyword arguments to pass to
                `Client.upload_fileobj`.

        Returns:
            The path that the folder was uploaded to.

        Examples:
            Upload contents from my_folder to new_folder.
            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            s3_bucket.upload_from_folder("my_folder", "new_folder")
            ```
        """
        from_folder = Path(from_folder)
        bucket_folder = self._join_bucket_folder(to_folder or "")

        num_uploaded = 0
        client = self.credentials.get_s3_client()

        async_coros = []
        for from_path in from_folder.rglob("**/*"):
            # this skips the actual directory itself, e.g.
            # `my_folder/` will be skipped
            # `my_folder/notes.txt` will be uploaded
            if from_path.is_dir():
                continue
            bucket_path = (
                Path(bucket_folder) / from_path.relative_to(from_folder)
            ).as_posix()
            self.logger.info(
                f"Uploading from {str(from_path)!r} to the bucket "
                f"{self.bucket_name!r} path {bucket_path!r}."
            )
            async_coros.append(
                run_sync_in_worker_thread(
                    client.upload_file,
                    Filename=str(from_path),
                    Bucket=self.bucket_name,
                    Key=bucket_path,
                    **upload_kwargs,
                )
            )
            num_uploaded += 1
        await asyncio.gather(*async_coros)

        if num_uploaded == 0:
            self.logger.warning(f"No files were uploaded from {str(from_folder)!r}.")
        else:
            self.logger.info(
                f"Uploaded {num_uploaded} files from {str(from_folder)!r} to "
                f"the bucket {self.bucket_name!r} path {bucket_path!r}"
            )

        return to_folder

    @sync_compatible
    async def copy_object(
        self,
        from_path: Union[str, Path],
        to_path: Union[str, Path],
        to_bucket: Optional[Union["S3Bucket", str]] = None,
        **copy_kwargs,
    ) -> str:
        """Uses S3's internal
        [CopyObject](https://docs.aws.amazon.com/AmazonS3/latest/API/API_CopyObject.html)
        to copy objects within or between buckets. To copy objects between buckets,
        `self`'s credentials must have permission to read the source object and write
        to the target object. If the credentials do not have those permissions, try
        using `S3Bucket.stream_from`.

        Args:
            from_path: The path of the object to copy.
            to_path: The path to copy the object to.
            to_bucket: The bucket to copy to. Defaults to the current bucket.
            **copy_kwargs: Additional keyword arguments to pass to
                `S3Client.copy_object`.

        Returns:
            The path that the object was copied to. Excludes the bucket name.

        Examples:

            Copy notes.txt from my_folder/notes.txt to my_folder/notes_copy.txt.

            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            s3_bucket.copy_object("my_folder/notes.txt", "my_folder/notes_copy.txt")
            ```

            Copy notes.txt from my_folder/notes.txt to my_folder/notes_copy.txt in
            another bucket.

            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            s3_bucket.copy_object(
                "my_folder/notes.txt",
                "my_folder/notes_copy.txt",
                to_bucket="other-bucket"
            )
            ```
        """
        s3_client = self.credentials.get_s3_client()

        source_path = self._resolve_path(Path(from_path).as_posix())
        target_path = self._resolve_path(Path(to_path).as_posix())

        source_bucket_name = self.bucket_name
        target_bucket_name = self.bucket_name
        if isinstance(to_bucket, S3Bucket):
            target_bucket_name = to_bucket.bucket_name
            target_path = to_bucket._resolve_path(target_path)
        elif isinstance(to_bucket, str):
            target_bucket_name = to_bucket
        elif to_bucket is not None:
            raise TypeError(
                "to_bucket must be a string or S3Bucket, not"
                f" {type(target_bucket_name)}"
            )

        self.logger.info(
            "Copying object from bucket %s with key %s to bucket %s with key %s",
            source_bucket_name,
            source_path,
            target_bucket_name,
            target_path,
        )

        s3_client.copy_object(
            CopySource={"Bucket": source_bucket_name, "Key": source_path},
            Bucket=target_bucket_name,
            Key=target_path,
            **copy_kwargs,
        )

        return target_path

    @sync_compatible
    async def move_object(
        self,
        from_path: Union[str, Path],
        to_path: Union[str, Path],
        to_bucket: Optional[Union["S3Bucket", str]] = None,
    ) -> str:
        """Uses S3's internal CopyObject and DeleteObject to move objects within or
        between buckets. To move objects between buckets, `self`'s credentials must
        have permission to read and delete the source object and write to the target
        object. If the credentials do not have those permissions, this method will
        raise an error. If the credentials have permission to read the source object
        but not delete it, the object will be copied but not deleted.

        Args:
            from_path: The path of the object to move.
            to_path: The path to move the object to.
            to_bucket: The bucket to move to. Defaults to the current bucket.

        Returns:
            The path that the object was moved to. Excludes the bucket name.

        Examples:

            Move notes.txt from my_folder/notes.txt to my_folder/notes_copy.txt.

            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            s3_bucket.move_object("my_folder/notes.txt", "my_folder/notes_copy.txt")
            ```

            Move notes.txt from my_folder/notes.txt to my_folder/notes_copy.txt in
            another bucket.

            ```python
            from prefect_aws.s3 import S3Bucket

            s3_bucket = S3Bucket.load("my-bucket")
            s3_bucket.move_object(
                "my_folder/notes.txt",
                "my_folder/notes_copy.txt",
                to_bucket="other-bucket"
            )
            ```
        """
        s3_client = self.credentials.get_s3_client()

        source_path = self._resolve_path(Path(from_path).as_posix())
        target_path = self._resolve_path(Path(to_path).as_posix())

        source_bucket_name = self.bucket_name
        target_bucket_name = self.bucket_name
        if isinstance(to_bucket, S3Bucket):
            target_bucket_name = to_bucket.bucket_name
            target_path = to_bucket._resolve_path(target_path)
        elif isinstance(to_bucket, str):
            target_bucket_name = to_bucket
        elif to_bucket is not None:
            raise TypeError(
                "to_bucket must be a string or S3Bucket, not"
                f" {type(target_bucket_name)}"
            )

        self.logger.info(
            "Moving object from s3://%s/%s to s3://%s/%s",
            source_bucket_name,
            source_path,
            target_bucket_name,
            target_path,
        )

        # If invalid, should error and prevent next operation
        s3_client.copy(
            CopySource={"Bucket": source_bucket_name, "Key": source_path},
            Bucket=target_bucket_name,
            Key=target_path,
        )
        s3_client.delete_object(Bucket=source_bucket_name, Key=source_path)
        return target_path



from unittest.mock import MagicMock

import pytest
from moto import mock_ec2
from prefect import flow

from prefect_aws.client_waiter import client_waiter


@pytest.fixture
def mock_waiter(monkeypatch):
    waiter = MagicMock()
    monkeypatch.setattr(
        "prefect_aws.client_waiter.create_waiter_with_client",
        waiter,
    )
    return waiter


@pytest.fixture
def mock_client(monkeypatch, mock_waiter):
    client_mock = MagicMock(
        waiter_names=["instance_exists"], get_waiter=lambda waiter_name: mock_waiter
    )
    client_creator_mock = MagicMock(
        ClientCreator=lambda *args, **kwargs: MagicMock(
            create_client=lambda *args, **kwargs: client_mock
        )
    )
    monkeypatch.setattr("botocore.client", client_creator_mock)
    return client_creator_mock


@mock_ec2
def test_client_waiter_custom(mock_waiter, aws_credentials):
    @flow
    def test_flow():
        waiter = client_waiter(
            "batch",
            "JobExists",
            aws_credentials,
            waiter_definition={"waiters": {"JobExists": ["definition"]}, "version": 2},
        )
        return waiter

    test_flow()
    assert mock_waiter.wait.called_once_with("JobExists")


@mock_ec2
def test_client_waiter_custom_no_definition(mock_waiter, aws_credentials):
    @flow
    def test_flow():
        waiter = client_waiter("batch", "JobExists", aws_credentials)
        return waiter

    with pytest.raises(ValueError, match="The waiter name, JobExists"):
        test_flow()


@mock_ec2
def test_client_waiter_boto(mock_waiter, mock_client, aws_credentials):
    @flow
    def test_flow():
        waiter = client_waiter("ec2", "instance_exists", aws_credentials)
        return waiter

    test_flow()
    assert mock_waiter.wait.called_once_with("instance_exists")



"""
Prefect deployment steps for code storage and retrieval in S3 and S3
compatible services.
"""
from pathlib import Path, PurePosixPath
from typing import Dict, Optional

import boto3
from botocore.client import Config
from prefect._internal.compatibility.deprecated import deprecated_callable
from prefect.utilities.filesystem import filter_files, relative_path_to_current_platform
from typing_extensions import TypedDict


class PushToS3Output(TypedDict):
    """
    The output of the `push_to_s3` step.
    """

    bucket: str
    folder: str


@deprecated_callable(start_date="Jun 2023", help="Use `PushToS3Output` instead.")
class PushProjectToS3Output(PushToS3Output):
    """Deprecated. Use `PushToS3Output` instead."""


class PullFromS3Output(TypedDict):
    """
    The output of the `pull_from_s3` step.
    """

    bucket: str
    folder: str
    directory: str


@deprecated_callable(start_date="Jun 2023", help="Use `PullFromS3Output` instead.")
class PullProjectFromS3Output(PullFromS3Output):
    """Deprecated. Use `PullFromS3Output` instead.."""


@deprecated_callable(start_date="Jun 2023", help="Use `push_to_s3` instead.")
def push_project_to_s3(*args, **kwargs):
    """Deprecated. Use `push_to_s3` instead."""
    push_to_s3(*args, **kwargs)


def push_to_s3(
    bucket: str,
    folder: str,
    credentials: Optional[Dict] = None,
    client_parameters: Optional[Dict] = None,
    ignore_file: Optional[str] = ".prefectignore",
) -> PushToS3Output:
    """
    Pushes the contents of the current working directory to an S3 bucket,
    excluding files and folders specified in the ignore_file.

    Args:
        bucket: The name of the S3 bucket where files will be uploaded.
        folder: The folder in the S3 bucket where files will be uploaded.
        credentials: A dictionary of AWS credentials (aws_access_key_id,
            aws_secret_access_key, aws_session_token).
        client_parameters: A dictionary of additional parameters to pass to the boto3
            client.
        ignore_file: The name of the file containing ignore patterns.

    Returns:
        A dictionary containing the bucket and folder where files were uploaded.

    Examples:
        Push files to an S3 bucket:
        ```yaml
        push:
            - prefect_aws.deployments.steps.push_to_s3:
                requires: prefect-aws
                bucket: my-bucket
                folder: my-project
        ```

        Push files to an S3 bucket using credentials stored in a block:
        ```yaml
        push:
            - prefect_aws.deployments.steps.push_to_s3:
                requires: prefect-aws
                bucket: my-bucket
                folder: my-project
                credentials: "{{ prefect.blocks.aws-credentials.dev-credentials }}"
        ```

    """
    s3 = get_s3_client(credentials=credentials, client_parameters=client_parameters)

    local_path = Path.cwd()

    included_files = None
    if ignore_file and Path(ignore_file).exists():
        with open(ignore_file, "r") as f:
            ignore_patterns = f.readlines()

        included_files = filter_files(str(local_path), ignore_patterns)

    for local_file_path in local_path.expanduser().rglob("*"):
        if (
            included_files is not None
            and str(local_file_path.relative_to(local_path)) not in included_files
        ):
            continue
        elif not local_file_path.is_dir():
            remote_file_path = Path(folder) / local_file_path.relative_to(local_path)
            s3.upload_file(
                str(local_file_path), bucket, str(remote_file_path.as_posix())
            )

    return {
        "bucket": bucket,
        "folder": folder,
    }


@deprecated_callable(start_date="Jun 2023", help="Use `pull_from_s3` instead.")
def pull_project_from_s3(*args, **kwargs):
    """Deprecated. Use `pull_from_s3` instead."""
    pull_from_s3(*args, **kwargs)


def pull_from_s3(
    bucket: str,
    folder: str,
    credentials: Optional[Dict] = None,
    client_parameters: Optional[Dict] = None,
) -> PullFromS3Output:
    """
    Pulls the contents of an S3 bucket folder to the current working directory.

    Args:
        bucket: The name of the S3 bucket where files are stored.
        folder: The folder in the S3 bucket where files are stored.
        credentials: A dictionary of AWS credentials (aws_access_key_id,
            aws_secret_access_key, aws_session_token).
        client_parameters: A dictionary of additional parameters to pass to the
            boto3 client.

    Returns:
        A dictionary containing the bucket, folder, and local directory where
            files were downloaded.

    Examples:
        Pull files from S3 using the default credentials and client parameters:
        ```yaml
        pull:
            - prefect_aws.deployments.steps.pull_from_s3:
                requires: prefect-aws
                bucket: my-bucket
                folder: my-project
        ```

        Pull files from S3 using credentials stored in a block:
        ```yaml
        pull:
            - prefect_aws.deployments.steps.pull_from_s3:
                requires: prefect-aws
                bucket: my-bucket
                folder: my-project
                credentials: "{{ prefect.blocks.aws-credentials.dev-credentials }}"
        ```
    """
    s3 = get_s3_client(credentials=credentials, client_parameters=client_parameters)

    local_path = Path.cwd()

    paginator = s3.get_paginator("list_objects_v2")
    for result in paginator.paginate(Bucket=bucket, Prefix=folder):
        for obj in result.get("Contents", []):
            remote_key = obj["Key"]

            if remote_key[-1] == "/":
                # object is a folder and will be created if it contains any objects
                continue

            target = PurePosixPath(
                local_path
                / relative_path_to_current_platform(remote_key).relative_to(folder)
            )
            Path.mkdir(Path(target.parent), parents=True, exist_ok=True)
            s3.download_file(bucket, remote_key, str(target))

    return {
        "bucket": bucket,
        "folder": folder,
        "directory": str(local_path),
    }


def get_s3_client(
    credentials: Optional[Dict] = None,
    client_parameters: Optional[Dict] = None,
) -> dict:
    if credentials is None:
        credentials = {}
    if client_parameters is None:
        client_parameters = {}

    # Get credentials from credentials (regardless if block or not)
    aws_access_key_id = credentials.get("aws_access_key_id", None)
    aws_secret_access_key = credentials.get("aws_secret_access_key", None)
    aws_session_token = credentials.get("aws_session_token", None)

    # Get remaining session info from credentials, or client_parameters
    profile_name = credentials.get(
        "profile_name", client_parameters.get("profile_name", None)
    )
    region_name = credentials.get(
        "region_name", client_parameters.get("region_name", None)
    )

    # Get additional info from client_parameters, otherwise credentials input (if block)
    aws_client_parameters = credentials.get("aws_client_parameters", client_parameters)
    api_version = aws_client_parameters.get("api_version", None)
    endpoint_url = aws_client_parameters.get("endpoint_url", None)
    use_ssl = aws_client_parameters.get("use_ssl", True)
    verify = aws_client_parameters.get("verify", None)
    config_params = aws_client_parameters.get("config", {})
    config = Config(**config_params)

    session = boto3.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        profile_name=profile_name,
        region_name=region_name,
    )
    return session.client(
        "s3",
        api_version=api_version,
        endpoint_url=endpoint_url,
        use_ssl=use_ssl,
        verify=verify,
        config=config,
    )
